import torch
import torch.nn as nn

# def ISSchNet
def shifted_softplus(x):
    r"""Compute shifted soft-plus activation function.
    .. math::
       y = \ln\left(1 + e^{-x}\right) - \ln(2)
    Args:
        x (torch.Tensor): input tensor.
    Returns:
        torch.Tensor: shifted soft-plus of input.
        
    p.s. from SchNetPack
    """
    return nn.functional.softplus(x) - float(0.6931472)

class _PairDistFunc_norm(torch.autograd.Function):
    @staticmethod
    def forward(ctx, rs):
        r_ij = rs[..., None, :] - rs[..., None, :, :]
        d_ij = torch.norm(r_ij, 2, -1)
        ctx.save_for_backward(rs, r_ij, d_ij)
        return d_ij

    @staticmethod
    def backward(ctx, dy):
        rs, r_ij, d_ij = ctx.saved_tensors
        matrix = dy[..., None] * r_ij / (d_ij[..., None] + 1e-6)
        return torch.sum(matrix, -2) - torch.sum(matrix, -3)
pair_dist = _PairDistFunc_norm.apply


def pair_dist_4_simu(rs):
    # implemented with torch.norm to get rid of gradient problem
    r_ij = rs[..., None, :] - rs[..., None, :, :]
    d_ij = torch.norm(r_ij, 2, -1)
    return d_ij

class RBF(nn.Module):
    """Radial Basis (Gaussian) Functions. Not trainable.
    Shared among all interaction layers."""
    def __init__(self, N_kernels=50, cutoff=5, gamma=0.1, training=False):
        super(RBF, self).__init__()
        self._mus = nn.Parameter(torch.arange(0, cutoff, cutoff / N_kernels), requires_grad=False)
        self._ngamma = nn.Parameter(torch.tensor(-gamma), requires_grad=False)
        if training:
            self._pair_dist_func = pair_dist
        else:
            self._pair_dist_func = pair_dist_4_simu

    def forward(self, rs):
        # rs: Tensor(batch, N_atoms, 3): atomic coords
        #d_ij = torch.cdist(rs, rs) # problem: no backward
        #d_ij = torch.sqrt(1e-6 + torch.sum(torch.pow(rs[:, :, None, :] - rs[:, None, :, :], 2), -1))
        d_ij = self._pair_dist_func(rs)
        # d_ij: Tensor(batch, N_atoms, N_atoms): interatomic dists
        return torch.exp(self._ngamma * torch.pow(d_ij[..., None] - self._mus, 2))

def charge_rbf(charges, N_kernels=40, gamma=50.):
    mus = torch.linspace(-1, 1, N_kernels + 1)[:N_kernels]
    return torch.exp(-gamma * torch.square(charges[..., None] - mus))

class FilterGen(nn.Module):
    # PBC pooling omitted
    def __init__(self, N_RBF_kernels=50, width=64, activation=shifted_softplus):
        super(FilterGen, self).__init__()
        self.fc1 = nn.Linear(N_RBF_kernels, width)
        self.fc2 = nn.Linear(width, width)
        self._act = activation

    def forward(self, rbf_output):
        # rbf_output: Tensor(batch, N_atoms, N_atoms, #RBF)
        y = self._act(self.fc1(rbf_output))
        # filters: Tensor(batch, N_atoms, N_atoms, width)
        return self._act(self.fc2(y))

class InteractionBlock(nn.Module):
    def __init__(self, width=64, N_RBF_kernels=50, activation=shifted_softplus, filter_gen_activation=shifted_softplus):
        super(InteractionBlock, self).__init__()
        self.pre_aw = nn.Linear(width, width)
        self.filter_gen = FilterGen(N_RBF_kernels, width, filter_gen_activation)
        self.post_aw1 = nn.Linear(width, width)
        self.post_aw2 = nn.Linear(width, width)
        self._act = activation
    
    def forward(self, x, rbf_output):
        # x: Tensor(batch, N_atoms, width): from an embedding layer or interaction layer
        # rbf_output: Tensor(batch, N_atoms, N_atoms)
        y = self.pre_aw(x)
        # cfconv
        y = torch.sum(y[:, None, :, :] * self.filter_gen(rbf_output), 2)
        y = self._act(self.post_aw1(y))
        return x + self.post_aw2(y)

class ISSchNet(nn.Module):
    # this is a unified version of different types of embeddings
    def __init__(self, mol_setup, N_ib=1, type_emb_width=32, charge_emb_width=0, N_RBF_kernels=32, RBF_cutoff=5, RBF_gamma=.1, act_func={"default":shifted_softplus}):
        super(ISSchNet, self).__init__()
        atom_types, charges = mol_setup
        self.N_atom = len(atom_types)
        if type_emb_width > 0:
            atom_types = torch.tensor(atom_types)
            self.Z = nn.Parameter(atom_types, requires_grad=False)
            self.type_emb = nn.Embedding(8, type_emb_width)
        if charge_emb_width > 0:
            charges = torch.tensor(charges, requires_grad=False, dtype=torch.float32)
            self.charge_rbf = nn.Parameter(charge_rbf(charges, N_kernels=40), requires_grad=False)
            self.charge_emb = FilterGen(40, charge_emb_width)
        
        self.rbf = RBF(N_RBF_kernels, RBF_cutoff, RBF_gamma, training=True)

        if "filter_gen" not in act_func.keys():
            act_func["filter_gen"] = act_func["default"]
        if "interaction_block" not in act_func.keys():    
            act_func["interaction_block"] = act_func["default"]
        if "post_processing" not in act_func.keys():
            act_func["post_processing"] = act_func["default"]

        self.ibs = nn.ModuleList([InteractionBlock(type_emb_width + charge_emb_width, N_RBF_kernels, act_func["interaction_block"], act_func["filter_gen"]) for _ in range(N_ib)])
        self.aw = nn.Linear(type_emb_width + charge_emb_width, 32)
        self._act = act_func["post_processing"]
        self.atom_energy = nn.Linear(32, 1, bias=False)

    def _energy(self, coords):
        #y = self.emb(self.Z)[None, :, :]
        if hasattr(self, 'type_emb') and hasattr(self, 'charge_emb'):
            y = torch.cat([self.type_emb(self.Z), self.charge_emb(self.charge_rbf)], dim=1)[None, :, :]
        elif hasattr(self, 'type_emb'):
            y = self.type_emb(self.Z)[None, :, :]
        elif hasattr(self, 'charge_emb'):
            y = self.charge_emb(self.charge_rbf)[None, :, :]
        rbf_output = self.rbf(coords)
        for ib in self.ibs:
            y = ib(y, rbf_output)
        y = self._act(self.aw(y))
        energy = torch.sum(self.atom_energy(y), (2, 1))
        return energy
    
    def forward(self, x):
        # x should require grad
        energy = self._energy(x)
        forces = torch.autograd.grad(-torch.sum(energy), x,
                                    create_graph=True,
                                    retain_graph=True)[0]
        return forces

class ISSchNet_batchE(nn.Module):
    # batchE means this is a version that calc total energy for a batch of input conformations
    # can be scripted
    def __init__(self, mol_setup, N_ib=1, type_emb_width=32, charge_emb_width=0, N_RBF_kernels=32, RBF_cutoff=5, RBF_gamma=.1, act_func={"default":shifted_softplus}):
        super(ISSchNet_batchE, self).__init__()
        atom_types, charges = mol_setup
        self.N_atom = len(atom_types)
        if type_emb_width > 0:
            atom_types = torch.tensor(atom_types)
            self.Z = nn.Parameter(atom_types, requires_grad=False)
            self.type_emb = nn.Embedding(8, type_emb_width)
        if charge_emb_width > 0:
            charges = torch.tensor(charges, requires_grad=False, dtype=torch.float32)
            self.charge_rbf = nn.Parameter(charge_rbf(charges, N_kernels=40), requires_grad=False)
            self.charge_emb = FilterGen(40, charge_emb_width)
        
        self.rbf = RBF(N_RBF_kernels, RBF_cutoff, RBF_gamma, training=False)

        if "filter_gen" not in act_func.keys():
            act_func["filter_gen"] = act_func["default"]
        if "interaction_block" not in act_func.keys():    
            act_func["interaction_block"] = act_func["default"]
        if "post_processing" not in act_func.keys():
            act_func["post_processing"] = act_func["default"]

        self.ibs = nn.ModuleList([InteractionBlock(type_emb_width + charge_emb_width, N_RBF_kernels, act_func["interaction_block"], act_func["filter_gen"]) for _ in range(N_ib)])
        self.aw = nn.Linear(type_emb_width + charge_emb_width, 32)
        self._act = act_func["post_processing"]
        self.atom_energy = nn.Linear(32, 1, bias=False)

    def _energy(self, coords):
        #y = self.emb(self.Z)[None, :, :]
        if hasattr(self, 'type_emb') and hasattr(self, 'charge_emb'):
            y = torch.cat([self.type_emb(self.Z), self.charge_emb(self.charge_rbf)], dim=1)[None, :, :]
        elif hasattr(self, 'type_emb'):
            y = self.type_emb(self.Z)[None, :, :]
        elif hasattr(self, 'charge_emb'):
            y = self.charge_emb(self.charge_rbf)[None, :, :]
        rbf_output = self.rbf(coords)
        for ib in self.ibs:
            y = ib(y, rbf_output)
        y = self._act(self.aw(y))
        energy = torch.sum(self.atom_energy(y), (2, 1))
        return energy
    
    def forward(self, x):
        # x should require grad
        xx = x.reshape([-1, self.N_atom, 3]) * 10. # unit conv: nm (OpenMM) -> Angstrom (model)
        sum_energy = torch.sum(self._energy(xx))
        return sum_energy

