import numpy as np
import torch
import sys

import dataset_util 
from train_util import Model_CUDA
from ISSchNet import ISSchNet, shifted_softplus
from mol_setup import load_mol_setup

# preset parameter
N_folds = 4
dataset_dir = "./dataset/"
AA_name = "Ala2"

embedding_type = sys.argv[1] # can be t, q or qt
# q means we use atomic partial charge in embedding
# t means we use also atom types as embedding
fold_index = int(sys.argv[2]) # which fold in cross validation

config = {
        "size": 1e6,
        "lr": 0.001,
        "lr_decay": False,
        "epochs": 20,
        "N_ib": 2,
        "width": 32,
        "N_RBF_kernels": 32,
        "RBF_cutoff": 5.0,
        "act_func": {
            "filter_gen": "ssp",
            "interaction_block": "ssp",
            "post_processing": "ssp",
            "default": "ssp"
        },
        "output_file": "output/%s_%s" % (AA_name, embedding_type)
}

type_emb_width, charge_emb_width = 0, 0
if embedding_type == "t":
    type_emb_width = config["width"]
elif embedding_type == "q":
    charge_emb_width = config["width"]
else:
    type_emb_width = int(config["width"] / 2)
    charge_emb_width = config["width"] - type_emb_width

## define a print helper
from datetime import datetime
def printDT(*args, **kwargs):
    print(datetime.now().strftime("[%d/%m/%Y %H:%M:%S]"), *args, **kwargs)
printDT("Training setup:", config)

N_epochs = config["epochs"]
size = int(config["size"])
lr = config["lr"]
lr_decay = config["lr_decay"]

for (k, v) in config["act_func"].items():
    if v == "tanh":
        config["act_func"][k] = torch.tanh
    elif v == "ssp":
        config["act_func"][k] = shifted_softplus
    else:
        raise NotImplementedError


dataset = dataset_util.Ala2DatasetCV(dataset_dir, size, N_folds)

# initiate a ISSchnet model
mol_setup = load_mol_setup(dataset_dir + "capped_ala_amber99sb_mol_setup.npz")
solv = ISSchNet(mol_setup, config["N_ib"], type_emb_width, charge_emb_width, config["N_RBF_kernels"], config["RBF_cutoff"], act_func=config["act_func"]).cuda()
optimizer = torch.optim.Adam(solv.parameters(), lr=lr)
if lr_decay:
    opt_scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, 0.95)
criterion = torch.nn.MSELoss()
model = Model_CUDA(solv, optimizer, criterion, dataset.get_training_set(fold_index), dataset.get_validation_set(fold_index), batch_size_train=128)


train_losses = np.zeros(N_epochs + 1)
val_losses = np.zeros(N_epochs + 1)
val_losses[0] = model.val_loss()
train_losses[0] = np.NaN
best_val_loss = np.inf
for j in range(N_epochs):
    history = model.train_epoch()
    val_losses[j + 1] = history["val_loss"]
    train_losses[j + 1] = history["train_loss"]
    if(lr_decay):
        opt_scheduler.step()
    if(val_losses[j + 1] < best_val_loss):
        best_val_loss = val_losses[j + 1]
        torch.save(solv.state_dict(), config["output_file"] + "_weights_%d.pt" % fold_index)
        torch.save(optimizer.state_dict(), config["output_file"] + "_opt_status_%d.pt" % fold_index)
    printDT("Epoch %d ends." % j)
    np.savez(config["output_file"] + "_losses_%d.npz" % (fold_index), val_losses=val_losses, train_losses=train_losses)
printDT("Training ends.")

